from .primitive import PrimitiveBasis
from .legendre import LegendreBasis

__all__ = ['PrimitiveBasis', 'LegendreBasis']

DiscreteIndicatorBasis = PrimitiveBasis
ContinuousIndicatorBasis = PrimitiveBasis
